/** 
 *   Camera Control for STAR I.
 *
 *   (c) Copyright 2005 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include "my_stdio.h"

#include "camera.h"
#include "fits_io.h"
#include "camera_tools.h"
#include "parse_tools.h"
#include "version.h"
#include "novas.h"
#include "ang_tools.h"

char *local_config_file = NULL;
static int set_location_flag = 0;

#define DEFAULT_EQUINOX 2000.0

site_info geo_loc = {
	.latitude = 0.0,
	.longitude = 0.0,
	.height = 0.0,
	.temperature = -273.0,
	.pressure = 0.0,
};

int warn_temp = 0;
int warn_temp_flag = 0;


int set_location(void)
{
	struct image_2d im;
	int ret;
	char cmd[_POSIX_PATH_MAX];
	char *editor;

	local_config_file = (char *) malloc(_POSIX_PATH_MAX * sizeof(char));
	TEST_MALLOC(local_config_file, "local_config_file");

	sprintf(local_config_file,"%s/%s", getenv("HOME"), CONFIG_FILE);
	im.ft = (struct fit_tag *) malloc(sizeof(struct fit_tag));
	TEST_MALLOC(im.ft, "struct fit_tag");
	im.ft->next = NULL;
	ct_read_header_file(&im, GLOBAL_CONFIG_FILE);
	do {
		ret = ct_read_header_file(&im, local_config_file);
		if (ret == -1) {
			sprintf(cmd, "cp %s %s", USER_DEF_CONFIG_FILE,
				local_config_file);
			system(cmd);
			editor = getenv("EDITOR");
			if( editor == NULL) editor = DEF_EDITOR;
			sprintf(cmd,"%s  %s", editor, local_config_file);
			system(cmd);
		}
	} while(ret == -1);
	geo_loc.latitude = im.latitude;
	geo_loc.longitude = im.longitude;
	geo_loc.height = im.altitude;
	geo_loc.temperature = 20.0; /* not used */
	geo_loc.pressure = 1020.0; /* not used */
	fits_tag_free(im.ft);

	set_location_flag = 1;
	return 0;
}


static double tm2julian_date(struct tm *t)
{
	return julian_date((short int)(t->tm_year + 1900), 
			   (short int)(t->tm_mon + 1), 
			   (short int)t->tm_mday, 
			   (double)t->tm_hour + 
			   (((double)t->tm_sec)/60.0 + 
			    ((double)t->tm_min))/60.0);
}


int get_object_position(double *ra, double *dec)
{
	
	struct image_2d im;
	struct tm *t;
	struct timeval tv;
	int ret = -1;

	if (!set_location_flag) {
		PVERB(PERR,"No location setting: abort\n");
		return -1;
	}

	im.ft = (struct fit_tag *) malloc(sizeof(struct fit_tag));
	TEST_MALLOC(im.ft, "struct fit_tag");
	im.ft->next = NULL;
	im.read_flag = 0;
	ct_read_header_file(&im, local_config_file);
	if ( (im.read_flag & (IMAGE_2D_RA|IMAGE_2D_DEC|IMAGE_2D_EQUINOX)) == 
	     (IMAGE_2D_RA|IMAGE_2D_DEC|IMAGE_2D_EQUINOX) ) {
		gettimeofday(&tv, NULL);
		t = gmtime( (time_t *) &tv.tv_sec);
		im.tjd = tm2julian_date(t);
		ct_precess(im.equinox, im.tjd, 
			   strdeg2double(im.ra_cat),
			   strdeg2double(im.dec_cat), ra, dec);
		ret = 0;
	} else {
		*ra = *dec = 0.0;
		ret = -(im.read_flag|IMAGE_2D_ERROR);
	}

	fits_tag_free(im.ft);

	return ret;
}


int ct_read_image_and_info(int cam, struct image_2d **image)
{
	int len ;
	int i, k;
	int *data;
	char str_tmp[80];
	struct image_2d *im;
	struct tm *t;
	double rar, decr;
	double ra1, dec1;
	
	if (!set_location_flag) {
		PVERB(PERR,"No location setting: abort\n");
		return CT_IM_LOC_SET;
	}

	im = *image = (struct image_2d *) malloc(sizeof(struct image_2d));
	TEST_MALLOC(im, "struct image_2d");
	im->read_flag = 0;
	im->equinox = DEFAULT_EQUINOX;
	im->ft = (struct fit_tag *) malloc(sizeof(struct fit_tag));
	TEST_MALLOC(im->ft, "struct fit_tag");
	im->ft->next = NULL;
	im->data = NULL;

	/* FIXME: cc_act_rcv_setup() no resetea el valore de binning */
	if ( cc_act_rcv_setup(cam, &im->s) < 0) {
		PVERB(PWARN, " Error reading image setup.\n");
		return CT_IM_SETUP;
	}
	PVERB(PINFO, "binning: %d\n", im->s.binning);
	if (im->s.binning > 1) {
		switch (im->s.binning) {
			case 1: break;
			case 2: im->s.naxis1 = 288; break;
			case 3: im->s.naxis1 = 192; break;
			case 4: im->s.naxis1 = 144; break;
			case 5: im->s.naxis1 = 114; break;
			default:
				PVERB(PERR, "Binning factor ignored");
				break;
		}
		cc_set_box_x0(cam, (CCD_X_LEN - im->s.naxis1));
		cc_set_box_x_len(cam,  im->s.naxis1);
	}
	PVERB(PINFO, " *** dimension %d x %d\n", im->s.naxis2, im->s.naxis1);
	len = im->s.naxis2 * im->s.naxis1;
	if ( im->s.naxis2 > CCD_Y_LEN || im->s.naxis2 <= 0 ||  /* 384 */
	     im->s.naxis1 > CCD_X_LEN || im->s.naxis1 <= 0 ) { /* 576 */
		PVERB(PWARN, " invalid dimension %d x %d data image\n",
		      im->s.naxis2, im->s.naxis1);
		return CT_IM_DIM;
	}
	/* FIXME: mantengo un bloque de memoria con el tamanno maximo
	 * de imagen: */
	//data = (int *) malloc(len * sizeof(int) );
	data = (int *) malloc(CCD_Y_LEN * CCD_X_LEN * sizeof(int) );
	TEST_MALLOC(data, " data array");
	PVERB(PINFO, " geting %d x %d data image: ", 
	      im->s.naxis2, im->s.naxis1);
	if ( cc_act_rcv_data(cam, data, &len) < 0) {
		PVERB(PERR, " Error reading image.\n");
		free(data);
		return CT_IM_READ;
	}
	/* resetear la caja: FIXME: test*/
	cc_set_box_x0(cam, 0);
	cc_set_box_x_len(cam, CCD_X_LEN);
	
	im->data = data;
	im->datamin = *(data+2);
	im->datamax = *data;
	im->max_count = 1;
	for (i=1;i<len;i++) {
		/* Skip over scan region: */
		if( (i % im->s.naxis2) != 0 && ((i-1) % im->s.naxis2) != 0 ) {
			if( *(data+i) < im->datamin) im->datamin = *(data+i);
			else if(*(data+i) > im->datamax) {
				im->datamax = *(data+i);
				im->max_count = 1;
			}else if (*(data+i) == im->datamax) im->max_count++;
		}
	}
	/* ! FIXME: Recomponer la geometria en caso de binning: */
	if (im->s.binning > 1) {
		switch (im->s.binning) {
		case 1: break;
		case 2: im->s.naxis2 = 192;
			for (i=0,k=0; i<len; i++, k+=2) {
				*(data+i) = (*(data+k) + *(data+k+1)) >> 1; 
			}
			break;
		case 3: im->s.naxis2 = 128;
			for (i=0,k=0; i<len; i++, k+=3) {
				*(data+i) = (*(data+k) + 
					     *(data+k+1) + 
					     *(data+k+2) ) / 3.0; 
			}
			break;
		case 4: im->s.naxis2 = 96;
			for (i=0,k=0; i<len; i++, k+=4) {
				*(data+i) = (*(data+k) + 
					     *(data+k+1) + 
					     *(data+k+2) +
					     *(data+k+3) ) >> 2; 
			}
			break;
		case 5: im->s.naxis2 = 76; /* y sobran 4 pixeles. */
			for (i=0,k=0; i<len; i++, k+=5) {
				*(data+i) = (*(data+k) + 
					     *(data+k+1) + 
					     *(data+k+2) +
					     *(data+k+3) +
					     *(data+k+4) ) / 5.0;
				if (i % im->s.naxis2 == 0) {
					k+=4;
				}
			}
			break;
		default:
			PVERB(PERR, "Binning factor ignored");
			break;
		}
		len = im->s.naxis2 * im->s.naxis1;
	}
	PVERB(PINFO, " *** dimension %d x %d\n", im->s.naxis2, im->s.naxis1);
	sprintf(str_tmp,
		"stari camera control, (c) Federico A. Bareilles (License GPL)");
	comment_push(im->ft, str_tmp);
	fits_tag_push(im->ft, "SOFT_VER", VERSION, "Software version number");
	fits_tag_push(im->ft, "SYSUSER", getenv("USER"), "System user name");
	/* Aca contemple la inicializacionde RA & DEC inicialmente, pero... */
	/* im->ra_cat[0] = im->dec_cat[0] = 0; */
	sprintf(im->filter, "-");
	ct_read_header_file(im, GLOBAL_CONFIG_FILE);
	ct_read_header_file(im, local_config_file);

	/* Time and coordinates calculation: */
	t = gmtime( (time_t *) &im->s.tv.tv_sec);
	sprintf(im->date_obs,"%04d-%02d-%02d", t->tm_year + 1900, 
		t->tm_mon + 1, t->tm_mday);
	sprintf(im->time_obs,"%02d:%02d:%02d", t->tm_hour, 
		t->tm_min, t->tm_sec);
	im->tjd = tm2julian_date(t);
	sidereal_time(im->tjd, 0.0, 0.0, &im->gst);
	im->lst = mod24(im->gst + im->longitude / 15.0);

	if ((im->read_flag & (IMAGE_2D_RA|IMAGE_2D_DEC|IMAGE_2D_EQUINOX)) == 
	    (IMAGE_2D_RA|IMAGE_2D_DEC|IMAGE_2D_EQUINOX) ) {


		ct_precess(im->equinox, im->tjd, 
			   strdeg2double(im->ra_cat), 
			   strdeg2double(im->dec_cat), &ra1, &dec1);
		double2gms_str(im->ra, ra1);
		double2gms_str(im->dec, dec1);
		im->ha = mod24(im->lst - ra1);	
		equ2hor (im->tjd, DELTA_T, 0.0, 0.0,
			 &geo_loc,
			 ra1, dec1, 0,
			 &im->z, &im->azimuth, &rar, &decr);
		im->azimuth += 180.0;
		while(im->azimuth > 360.0) im->azimuth -= 360.0;
		im->airmas = airmas(im->z * _DEG2RAD);
	} else { /* RA, DEC or EQUINOX not defined */
		sprintf(im->ra_cat, "00:00:00");
		sprintf(im->dec_cat, "00:00:00");
		sprintf(im->ra, "00:00:00");
		sprintf(im->dec, "00:00:00");
		im->ha = 0;
		im->azimuth = 0;
		im->airmas = 0;
	}


	return CT_IM_OK;
}


/**
   This is the old function: 
   *) Never use this function in new software.
**/
#if 0
int ct_read_save_file(int cam, char *filename)
{

	int status;
	struct image_2d *im = NULL;

	if ((status = ct_read_image_and_info(cam, &im)) != CT_IM_OK) {
		image_2d_free(im);
		return status;
	}
	/* WARNING: this printf is only for compatibility... */
	printf("DATAMIN: %d, DATAMAX: %d (in %d pixels)\n",
	       im->datamin, im->datamax, im->max_count);

	status = image_2d_save(filename, im);
	image_2d_free(im);
	if (status != 0) {
		PVERB(PWARN, " Fail.\n");
		fits_report_error(stdout, status);
		return CT_IM_FITS_ERR;
	}

	return CT_IM_OK;
}
#endif


int ct_read_header_file(struct image_2d *im, char *filename)
{
	FILE *fp;
	char line[LINE_LEN_CONF];
	char *sline, *tag, *data, *comment;
	int i;

	fp = fopen(filename, "r");
	
	if( fp == NULL ){
                SYSLOG(LOG_ERR,
                       "ct_read_header_file: Error opening file: %s",
		       filename);
                return -1;
        }
	
	im->flat_type = FLAT_UNKNOW;
	memset(line, 0, sizeof(line));
	while( fgets(line, sizeof(line), fp)){
		if(line[strlen(line)-1] == '\n')
			line[strlen(line)-1] = 0;
		sline = stripwhite(line);
		if (parse_line(sline, &tag, &data, &comment) == 0){
			if( memcmp(tag,"OBJECT", 6) == 0 ){
				sprintf(im->object,"%s", data);
			}else if( memcmp(tag,"OBSERVER", 8) == 0 ){
				sprintf(im->observer, "%s", data);
			}else if( memcmp(tag,"OBSERVAT", 8) == 0 ){
				sprintf(im->observat, "%s", data);
			}else if( memcmp(tag,"TELESCOP", 8) == 0 ){
				sprintf(im->telescope, "%s", data);
			}else if( memcmp(tag,"INSTRUME", 8) == 0 ){
				sprintf(im->instrument, "%s", data);
			}else if( memcmp(tag,"PIXSCALE", 8) == 0 ){
				im->pixscale = atof(data);
			}else if( memcmp(tag,"RA", 2) == 0 ){
				sprintf(im->ra_cat, "%s", data);
				im->read_flag |= IMAGE_2D_RA;
			}else if( memcmp(tag,"DEC", 3) == 0 ){
				sprintf(im->dec_cat, "%s", data);
				im->read_flag |= IMAGE_2D_DEC;
			}else if( memcmp(tag,"EQUINOX", 7) == 0 ){
				im->equinox = atof(data);
				im->read_flag |= IMAGE_2D_EQUINOX;
			}else if( memcmp(tag,"FILTER", 6) == 0 ){
				for(i=0;i<strlen(data) && i < FLEN_VALUE;i++) {
					*(im->filter+i) = (char) 
						toupper(*(data+i));
				}
				if ( i < FLEN_VALUE ) *(im->filter+i) = 0;
				//sprintf(im->filter, "%s", data);
				
			}else if( memcmp(tag,"FLATTYPE", 8) == 0 ){
				switch(toupper(*(data))) {
				case 'S':
					im->flat_type = FLAT_SKY;
					break;
				case 'D':
					im->flat_type = FLAT_DOME;
					break;
				default:
					im->flat_type = FLAT_UNKNOW;
				}
			}else if( memcmp(tag,"LATITUDE", 8) == 0 ){
				im->latitude = strdeg2double(data);
			}else if( memcmp(tag,"LONGITUD", 8) == 0 ){
				im->longitude = strdeg2double(data);
			}else if( memcmp(tag,"ALTITUDE", 8) == 0 ){
				im->altitude = strdeg2double(data);
			}else if( memcmp(tag,"COMMENT", 7) == 0 ){
				comment_push(im->ft, data);
			}else{
				if(comment != NULL)
					fits_tag_push(im->ft, tag, data, 
						      comment);
				else
					fits_tag_push(im->ft, tag, data, "");
			}
		}
		memset(line, 0, sizeof(line));
	}
	fclose(fp);

	return 0;
}


int ct_precess(double from_date, double to_date,
	    double ra0,  double dec0,
	    double *ra1, double *dec1 )
{
	cat_entry star0 = {"SRC", "start", 1, ra0, dec0, 0.0, 0.0, 0.0, 0.0};
	cat_entry star1;

	transform_cat (3, 
		       from_date, &star0,
		       to_date, "nc", &star1);
	*ra1 = star1.ra;
	*dec1 = star1.dec;

	return 0;
}



int strisnum(char *str)
{
	int i;
	int l = strlen(str);

	for (i=0; i < l && (str+i) != NULL ; i++) {
		if ( ! (isdigit(*(str+i)) || *(str+i) == '.') ) {
			return -1;
		}
	}

	return 1;
}


int file_name_exist(char *filename, char *basename)
{
	struct stat dummy;
	int ret = -1;

	sprintf(filename,"%s", basename);
	if (stat(filename, &dummy) == 0) {
		ret = 1;
	} else {
		sprintf(filename,"%s.%s", basename, IMAGE_EXTENSION);
		if (stat(filename, &dummy) == 0) {
			ret = 1;
		} else {
			*(filename) = 0;
		}
	}

	return ret;
}


int get_filter(char *fil)
{
	int ret = -1;
	struct configuration *conf = NULL;
	struct configuration *cf_filter;

	read_conf_file(&conf, local_config_file);
	cf_filter = find_conf_tag(conf,"FILTER");
	if (cf_filter != NULL) {
		*fil = cf_filter->data[0];
		ret = 0;
	}
	free_conf(conf);

	return ret;
}


int set_filter(char fil)
{
	int ret;
	struct configuration *conf = NULL;
	char f[] = {fil, 0};

	read_conf_file(&conf, local_config_file);
	ret = change_conf_tag(conf, "FILTER", f);
	if (ret == 0) 
		ret = write_conf_file(conf, local_config_file);
	free_conf(conf);

	return ret;
}
